/* alixconf.c - reconfigure BIOS settings on ALIX2 motherboards
 * Willy Tarreau - 2009.
 * This code was derived from "flasher.c" provided by Pascal Dornier.
 * There was no original copyright, so consider this as public domain.
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <sys/io.h>

/* Big fat warning: if this code goes 64 bit it will need uint_32-like typedefs */

#define MSR_ID 0x1808
#define FLASH0_BASE 0xfff80000
#define FLASH0_CTL  0xffb80000
#define FLASH1_BASE 0xfff00000
#define FLASH1_CTL  0xffb00000
#define ROM_SIZE (512 * 1024)
#define PAGE_SIZE (64 * 1024)
#define PAGE_COUNT 8

#define ID_AMIC    0x09d37
#define ID_WINBOND 0x050da
#define ID_SST     0x050bf

const char *msrudev = "/dev/cpu/0/msr"; /* For UDEV */
const char *msrdev = "/dev/msr0"; /* For custom distro */

unsigned int jedec_id = 0;
int gpi_id = -1;
unsigned long flash_base = 0;
unsigned long flash_ctl = 0;
volatile unsigned char *mem_ptr = NULL;
volatile unsigned char *ctl_ptr = NULL;
unsigned char *img_ptr = NULL;
unsigned int mem_fd = -1;
unsigned long delay_factor = 1;
int need_update = 0;
int force_flag = 0;
int lpc_flag = 0;
int serial_div = 0;
int ide_mode = 0;


#ifdef WITHOUT_PERROR
inline void perror(const char *msg)
{
	fprintf(stderr, "%s\n", msg);
}
#endif

/* Delay functions - not likely to work on multicore, but same goes for RTDSC */
void delay_usec(int count)
{
	volatile unsigned long int i;
	for (i = 0; i < count * delay_factor; ++i);
}

void delay_calibrate(void)
{
	int count;
	unsigned long usec;
	
	count = 1000;
	while (1) {
		struct timeval start, end;

		gettimeofday(&start, 0);
		delay_usec(count);
		gettimeofday(&end, 0);
		usec = 1000000 * (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec);
		if (usec >= 250000)
			break;

		/* with enough resolution and some margin, we can directly get the proper value */
		if (usec > 5)
			count = count * 280000.0 / usec;
		else
			count = count * 2;
	}
	delay_factor = count / usec;
	/* printf("Delay calibration: %ldM loops per second, %d us for count %d\n",
	   (unsigned long)delay_factor,  usec, count); */
}

/* End delay functions */

/* ROM functions */
/* Lock the ROM for writing */

void rom_lock(int lock)
{
	unsigned int i;
	volatile unsigned char *ptr;
	for (i = 0; i <= 7; ++i)
	{
		ptr = ctl_ptr + (i * (unsigned int)0x10000) + (unsigned int)0x02;
		*ptr = (unsigned char)lock;
	}
}

/* Erase ROM page 0 - 7 */
int rom_zap(unsigned int page)
{
	volatile unsigned char *page_base;
	volatile unsigned int *check_ptr;
	int isblank;
	unsigned int i, retry_count, retry_delay;
	unsigned char first;
	unsigned char erase_cmd;
	
	page_base = mem_ptr + (page * PAGE_SIZE);
	/* Blank check */
	isblank = 1;
	check_ptr = (unsigned int *)page_base;
	for (i=0; i < (PAGE_SIZE / 4); ++i)
	{
		if ((*check_ptr) != 0xFFFFFFFF)
		{
			isblank = 0;
			break;
		}
		++check_ptr;
	}
	if (isblank)
	{
		return 1;
	}
	/* Select erase command and delays according to flash type */
	if ((jedec_id == ID_AMIC) ||
	    (jedec_id == ID_SST))
	{
		erase_cmd = 0x50;
		retry_count = 12500;
		retry_delay = 1000;
	} else
	{
		erase_cmd = 0x30;
		retry_count = 250;
		retry_delay = 50000;
	}
	/* Sector erase command */
	*(page_base + (unsigned int)0x5555) = (unsigned char)0xaa;
	*(page_base + (unsigned int)0x2aaa) = (unsigned char)0x55;
	*(page_base + (unsigned int)0x5555) = (unsigned char)0x80;
	*(page_base + (unsigned int)0x5555) = (unsigned char)0xaa;
	*(page_base + (unsigned int)0x2aaa) = (unsigned char)0x55;
	*(page_base + (unsigned int)0x5555) = (unsigned char)erase_cmd;
	first = *page_base;
	i = retry_count;
	while (i > 0)
	{
		delay_usec(retry_delay);
		first = *page_base;
		if (first == 0xFF)
		{
			break;
		}
		--i;
	}
	if (i == 0) /* Timeout, return error */
	{
		return 0;
	}
	return 1;
}

/* Write a ROM page 0 - 7 from buffer src */
int rom_write(unsigned int page, unsigned char *src)
{
	volatile unsigned char *page_base;
	volatile unsigned char *writecmd_ptr;
	volatile unsigned char *ptr;
	int i;
	unsigned int scratch;
	int retry, correct_count;
	unsigned char tmp;
	
	page_base = mem_ptr + (page * PAGE_SIZE);
	scratch = flash_base + (page * PAGE_SIZE);
	//printf("Page address: %x\n", scratch);
	scratch = scratch & (unsigned int)0xFFFF0000; // Hack
	//printf("Scratch after and: %x\n", scratch);
	scratch = scratch - flash_base;
	//printf ("Write command address: %x\n", scratch);
	writecmd_ptr = mem_ptr + scratch;
	//printf ("mem ptr: %p, Write command ptr: %p\n", mem_ptr, writecmd_ptr);
	
	ptr = page_base;
	for (i=0; i<PAGE_SIZE; ++i)
	{
		if ((*src) != 0xFF)
		{
			*(mem_ptr + (unsigned int)0x5555) = (unsigned char)0xaa;
			*(mem_ptr + (unsigned int)0x2aaa) = (unsigned char)0x55;
			*(mem_ptr + (unsigned int)0x5555) = (unsigned char)0xa0;
			*ptr = *src;
			retry = 1000;
			while (retry > 0)
			{
				correct_count = 0;
				/* Read three times, that's why ptr points to volatile */
				tmp = *ptr;
				if (tmp == *src) ++correct_count;
				tmp = *ptr;
				if (tmp == *src) ++correct_count;
				tmp = *ptr;
				if (tmp == *src) ++correct_count;
				/* All 3 read correctly ? */
				if (correct_count == 3) break;
				--retry;
			}
			if (retry == 0)
			{
				//printf("Flash write fail at count %d!\n", i);
				return 0;
			}
		}
		++src;
		++ptr;
	}
	return 1;
}

/* End ROM functions */

/* Enable/disable flash write */

int flash_write_set(int enable)
{
	int msr_fd;
	unsigned long long msr;
	unsigned long *msr_hi; /* Low 32 bits of 64 bit MSR */
	unsigned long tmp;
	off_t seek_res;
	
	msr_fd = open(msrudev, O_RDWR);
	if (msr_fd == -1) /* Fail silently and try to open it by the custom name */
	{
		msr_fd = open(msrdev, O_RDWR);
	}
	if (msr_fd == -1)
	{
		perror("Could not open msr device:");
		return 0;
	}
	seek_res = lseek(msr_fd, MSR_ID, SEEK_SET);
	if (seek_res == ((off_t) - 1))
	{
		perror("Could not seek to MSR:");
		close(msr_fd);
		return 0;
	}
	if (read(msr_fd, &msr, 8) != 8)
	{
		perror("Could not read MSR:");
		close(msr_fd);
		return 0;
	}
	msr_hi = (unsigned long *)(&msr);
	++msr_hi;
	//printf("msr before: %x\n", *msr_hi);
	tmp = 0x00ffffff;
	(*msr_hi) = (*msr_hi) & tmp;
	if (enable)
	{
		tmp = 0x21000000;
		(*msr_hi) = (*msr_hi) | tmp;
	}
	else
	{
		tmp = 0x25000000;
		(*msr_hi) = (*msr_hi) | tmp;
	}
	//printf("msr after: %x\n", *msr_hi);
	if (write(msr_fd, &msr, 8) != 8)
	{
		perror("Could not write MSR:");
		close(msr_fd);
		return 0;
	}
	read(msr_fd, &msr, 8);
	//printf("msr reread: %x\n", *msr_hi);
	close(msr_fd);
	return 1;
}

/* MMap flash, find chip type - id 0 is normal address, id 1 is secondary address */

int flash_map(int id)
{
	unsigned int gpi_bits = 0;
	switch(id)
	{
		case 0:
			flash_base = FLASH0_BASE;
			flash_ctl = FLASH0_CTL;
			break;
		case 1:
			flash_base = FLASH1_BASE;
			flash_ctl = FLASH1_CTL;
			break;
		default:
			return 0;
	}
	mem_fd = open("/dev/mem", O_RDWR);
	if (mem_fd == -1)
	{
		perror("Could not open /dev/mem");
		return 0;
	}
	mem_ptr = (unsigned char *) mmap(0, 512 * 1024, 
		   PROT_READ | PROT_WRITE, MAP_SHARED, mem_fd, flash_base);
	if (mem_ptr == MAP_FAILED)
	{
		perror("Could not map BIOS");
		close(mem_fd);
		return 0;
	}
	ctl_ptr = (unsigned char *) mmap(0, 512 * 1024, 
		   PROT_READ | PROT_WRITE, MAP_SHARED, mem_fd, flash_ctl);
	if (ctl_ptr == MAP_FAILED)
	{
		perror("Could not map BIOS control");
		munmap((void *)mem_ptr, 512 * 1024);
		close(mem_fd);
		return 0;
	}
	jedec_id = *((unsigned short *)(ctl_ptr + 0x40000));
	gpi_bits = *((unsigned char *)(ctl_ptr + 0x40100));
	if (gpi_bits <= 2)
	{
		gpi_id = gpi_bits;
	}
	/* Check if jedec_id makes sense */
	if ((jedec_id == ID_AMIC) ||
	    (jedec_id == ID_WINBOND) ||
	    (jedec_id == ID_SST))
	{
		return 1;
	}
	
	munmap((void *)mem_ptr, 512 * 1024);
	munmap((void *)ctl_ptr, 512 * 1024);
	close(mem_fd);
	return 0;
}

/* Try to MMap secondary address then primary */
int flash_map_any(void)
{
	if (lpc_flag)
	{
		if (flash_map(0))
		{
			gpi_id = 0; /* Force ID 0 for LPC */
			return 1;
		}
		return 0;
	}
	if (flash_map(1))
	{
		return 1;
	}
	else if (flash_map(0))
	{
		return 1;
	}
	return 0;
}

/* print usage with <cmd> as the program name, and exit with code <err> */
void usage(const char *cmd, int err)
{
	fprintf(stderr,
		"Usage: %s [args]:\n"
		"\t-h this help text\n"
		"\t-f force update even if same parameters\n"
		"\t-l force update LPC module from alix.0 !!!\n"
		"\t-s <speed> to change serial speed\n"
		"\t-i <mode> to change IDE mode (1=CHS, 2=LBA)\n",
		cmd);
	exit(err);
}

/* print <msg>, and exit with code <err> */
void die(const char *msg, int err)
{
	fprintf(stderr, "%s\n", msg);
	exit(err);
}

int main(int argc, char *argv[])
{
	unsigned char *data;
	int i;
	int serial_speed;
	int ret;

	for (i = 1; i < argc; ++i) {
		if (argv[i][0] != '-')
			break;

		if (argv[i][1] == '-') {
			i++;
			break;
		}

		switch (argv[i][1]) {
		case 'f':
			force_flag = 1;
			break;

		case 'l':
			lpc_flag = 1;
			break;

		case 's':
			if (argc <= i + 1)
				usage(argv[0], 1);

			serial_speed = atol(argv[i+1]);
			serial_div = 115200 / (serial_speed?:1);
			if (!serial_div || serial_div > 65535 || 115200/serial_div != serial_speed)
				die("Serial speed must be a divisor of 115200 greater than 1.\n", 1);
			i++;
			break;
		case 'i':
			if (argc <= i + 1)
				usage(argv[0], 1);

			ide_mode = atol(argv[i+1]); /* 0 = unchanged, 1 = CHS, 2+ = LBA */
			i++;
			break;
		case 'h':
			usage(argv[0], 0);
		default:
			usage(argv[0], 1);
		}
	}

	if (flash_map_any() == 0)
		die("Flash map fail, or unknown flash type!\n", 1);

	ret = 1;
	if (memcmp((void *)mem_ptr + 0x7FFD0, "PC Engines ALIX.2", 17) != 0) {
		fprintf(stderr, "PC Engines signature not found\n");
		goto out_unmap;
	}

	/* check that the config area (page 0) looks OK */
	for (i = 0x20; i < PAGE_SIZE; i++)
		if (mem_ptr[i] != 0xFF)
			break;

	if (i != PAGE_SIZE) {
		fprintf(stderr, "Page 0 does not look like a config area, refusing to proceed.\n");
		goto out_unmap;
	}

	if (!serial_div || mem_ptr[0x00] != serial_div) {
		/* do not report changes only if ask for same setting */
		printf("Currently configured serial speed: %d bauds\n",
		       mem_ptr[0x00] ? 115200/(int)mem_ptr[0x00] : 0);
	}

	if (!ide_mode || mem_ptr[0x02] != (ide_mode - 1)) {
		/* do not report changes only if ask for same setting */
		printf("Currently configured CHS/LBA: %s\n",
		       mem_ptr[0x02] ? "LBA" : "CHS");
	}

	need_update = 0;
	if (serial_div && mem_ptr[0x00] != serial_div)
		need_update = 1;

	if (ide_mode && mem_ptr[0x02] != (ide_mode - 1))
		need_update = 1;

	if (!need_update && !force_flag) {
		ret = 0;
		goto out_unmap;
	}

	printf("Calibrating delay loop... ");
	delay_calibrate();
	printf("done, factor=%ld.\n", delay_factor);

	if (flash_write_set(1) == 0) {
		fprintf(stderr, "Flash write setup fail!\n");
		goto out_unmap;
	}
	rom_lock(0);

	data = malloc(PAGE_SIZE);
	if (!data) {
		fprintf(stderr, "Memory allocation failed\n");
		goto out_unset;
	}

	memcpy(data, (void *)mem_ptr, PAGE_SIZE);

	if (serial_div)
		data[0] = serial_div;

	if (ide_mode)
		data[2] = ide_mode - 1;  /* LBA / CHS */

	if (rom_zap(0) == 0) {
		fprintf(stderr, "Failed to erase page 0!\n");
		goto out_free;
	}

	if (rom_write(0, data) == 0) {
		fprintf(stderr, "Failed to write page 0!\n");
		goto out_free;
	}
	printf("Config successfully updated!\n");

	/* out_unlock */
	rom_lock(1);
 out_free:
	free(data);
 out_unset:
	flash_write_set(0);
 out_unmap:
	munmap((void *)mem_ptr, ROM_SIZE);
	munmap((void *)ctl_ptr, ROM_SIZE);
	close(mem_fd);
	return ret;
}
